import unittest
from simpycity.core import Function, Query
from simpycity.model import SimpleModel
from simpycity import config
import psycopg2
from optparse import OptionParser
import sys

class handleTests(unittest.TestCase):
    
    pass


import psycopg2.extras
from simpycity import config as g_config
from contextlib import contextmanager

def d_out(text):

    if g_config.debug:
        print text


class Cursor(psycopg2.extras.DictCursor):
    """
    Add per-row callback option to standard cursor.
    """
    def __init__(self, *args, **kwargs):
        self.callback = kwargs.pop('callback', None)
        super(Cursor, self).__init__(*args, **kwargs)

    def fetchone(self):
        row = super(Cursor, self).fetchone()
        if self.callback:
            row = self.callback(row)
        return row

    def fetchall(self):
        rows = super(Cursor, self).fetchall()
        if self.callback:
            rows = [self.callback(_) for _ in rows]
        return rows

    def fetchmany(self, size=None):
        rows = super(Cursor, self).fetchmany(size)
        if self.callback:
            rows = [self.callback(_) for _ in rows]
        return rows

    def __iter__(self):
        res = super(Cursor, self).__iter__()
        while True:
            if self.callback:
                yield self.callback(res.next())
            else:
                yield res.next()


class TypedCursor(Cursor):
    """
    A cursor for result sets having only a single (typically composite) column.
    Rather than a row being a tuple, it is simply the value of the one column.
    """
    def execute(self, query, vars=None):
        super(TypedCursor, self).execute(query, vars)
        if len(self.description) != 1:
            raise Exception("Cursor must return exactly one column")

    def fetchone(self):
        row = super(TypedCursor, self).fetchone()
        if row and len(row) > 0:
            return row[0]
        else:
            return row

    def fetchall(self):
        rows = super(TypedCursor, self).fetchall()
        rows = [_[0] for _ in rows]
        return rows

    def fetchmany(self, size=None):
        rows = super(TypedCursor, self).fetchmany(size)
        rows = [_[0] for _ in rows]
        return rows

    def __iter__(self):
        res = super(TypedCursor, self).__iter__()
        while True:
            row = res.next()
            if row and len(row) > 0:
                yield row[0]
            else:
                yield row


class Handle(object):

    """
    A base Simpycity handle.
    a Handle is the wrapper object around a psycopg connection.
    """

    def __init__(self, dsn=None, config=None, isolation_level=None):

        self.conn = None

        self.config = config or g_config
        self.dsn = dsn or self.config.dsn()

        self.isolation_level = None
        if isolation_level is not None and isolation_level in [0,1,2]:
            self.isolation_level = isolation_level

        d_out("Handle.__init__: Creating DB connection")
        self.__reconnect__()
        d_out("Handle.__init__: Connection PID is %s" % self.conn.get_backend_pid() )

        if self.isolation_level is not None:
            self.conn.set_isolation_level(isolation_level)

    def __reconnect__(self):
        if self.conn and not self.conn.closed:
            self.close()

        self.conn = psycopg2.connect(self.dsn)

    def cursor(self,*args,**kwargs):
        d_out("Handle.cursor: Creating cursor..")
        if not self.open:
            raise Exception("Connection isn't open.")

        if 'cursor_factory' not in kwargs or kwargs['cursor_factory'] == None:
            kwargs["cursor_factory"] = Cursor
        callback = kwargs.pop('callback', None)
        cur = self.conn.cursor(*args,**kwargs)
        if callback:
            d_out('Handle.cursor() setting callback attrib {0}'.format(callback))
            cur.callback = callback
        return cur

    def execute(self, *args, **kwargs):
        return self.cursor().execute(*args, **kwargs)

    @property
    def autocommit(self):
        # We trust the user not to run SQL SET commands directly to
        # alter the value of default_transaction_isolation setting.
        return self.isolation_level == psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT

    def begin(self):
        d_out("Handle.begin: Begin transaction.")

        if not self.open:
            raise Exception("Connection isn't open.")

        if self.autocommit:
            # In non-autocommit mode, the first statement executed on
            # cursor is prepended with BEGIN automatically by
            # psycopg2.  We only need to run it in autocommit mode.
            self.execute("BEGIN")

    def commit(self):
        d_out("Handle.commit: Commit transaction.")

        if self.conn.closed:
            # That's weird, and bad.
            raise Exception("Attempting to commit a closed handle.")

        if self.autocommit:
            # Connection object's commit/rollback() funcs has no
            # effect in autocommit mode.  Run direct SQL statement
            # through the default cursor.
            self.execute("COMMIT")
        else:
            self.conn.commit()

    def rollback(self):
        d_out("Handle.rollback: Abort transaction.")

        if not self.conn.closed:

            if self.autocommit:
                # See commit()
                self.execute("ROLLBACK")
            else:
                self.conn.rollback()

    @contextmanager
    def transaction(self):
        try:
            self.begin()
            yield
            self.commit()
        except:
            self.rollback()
            raise

    def __repr__(self):
        if not self.conn.closed:
            return "Handle object: pid %s" % self.conn.get_backend_pid()
        else:
            return "Handle object: no pid (closed)"

    def close(self,*args,**kwargs):
        d_out("Handle.close: de-allocating connection" )

        if self.conn is None:
            return

        if not self.conn.closed:
            d_out("handle.close: handle open, closing pid %s" % self.conn.get_backend_pid() )
            self.conn.close()
        else:
            d_out("handle.close: handle already closed.")

    def __del__(self):
        d_out("Handle.__del__: destroying handle, de-allocating connection")
        if self.conn:
            self.close()

    @property
    def open(self):
        try:
            if self.conn.get_backend_pid():
                return True
        except psycopg2.InterfaceError, e:
            if str(e) == "connection already closed":
                # We already lost our connection. Attempt to reforge it.
                self.__reconnect__()
                if self.conn.get_backend_pid() >= 0:
                    return True # Resets it.
            else:
                raise e # Another error. Don't try to trap it.



import unittest
from simpycity import config
from simpycity.core import *
from simpycity.model import SimpleModel, Construct
from psycopg2.extensions import cursor as _cursor
import psycopg2

import ConfigParser

handle = None

def test_handle_factory(*args, **kwargs):

    global handle
    if not handle:
        print("test_handle_factory: make new handle")
        from simpycity.handle import Handle
        handle = Handle(*args, **kwargs)

    return handle


def setUpModule():

    cfg = ConfigParser.ConfigParser()
    ini = cfg.read("test.ini")
    try:
        assert(ini)
    except AssertionError:
        print("Run the tests from the test directory")
        raise
    config.database =   cfg.get("simpycity","database")
    config.port =       cfg.get("simpycity","port")
    config.user =       cfg.get("simpycity","user")
    config.host =       cfg.get("simpycity","host")
    config.password =   cfg.get("simpycity","password")
    config.debug =      cfg.getboolean("simpycity","debug")

    config.handle_factory = test_handle_factory

    # clean up state
    h = open("sql/test_unload.sql","r")
    destroy_sql = h.read()
    h.close()

    handle = config.handle_factory() #, isolation_level=0)
    with handle.transaction():
        try:
            handle.execute(destroy_sql)
        except psycopg2.ProgrammingError:
            pass

class dbTest(unittest.TestCase):

    def setUp(self):
        h = open("sql/test.sql","r")
        create_sql = h.read()
        h.close()

        handle = config.handle_factory()
        with handle.transaction():
            handle.execute(create_sql)
        handle.commit()


    def tearDown(self):
        handle = config.handle_factory()
        handle.rollback()
        #handle.close()

        h = open("sql/test_unload.sql","r")
        destroy_sql = h.read()
        h.close()

        with handle.transaction():
            handle.execute(destroy_sql)


class ConstructTest(dbTest):

    def testConstructFunction(self):
        class o(Construct):
            r = Raw("SELECT * FROM test_table")

        instance = o()
        cur = instance.r()
        self.failUnless(
            isinstance(cur, _cursor),
            "Construct function doers not return expected result type."
        )

    def testCreateConstruct(self):

        class o(Construct):
            r = Raw("SELECT * FROM test_table")

        instance = o()
        self.assertTrue(
            isinstance(instance,Construct),
            "Construct not created successfully."
        )

class ModelTest(dbTest):

    def testCreateModel(self):

        q = SimpleReturn()
        self.assertTrue(
            isinstance(q, SimpleModel),
            "Model object not successfully created."
        )

    def testInstanceMethods(self):
        q = SimpleInstanceModel()
        self.assertTrue(
            isinstance(q, SimpleModel),
            "Model object not successfully created."
        )


    def testInstanceLoader(self):
        q = SimpleLoaderModel(id=1)
        self.assertEqual(
            q.id,
            1,
            "Model id is set to 1."
        )
        self.assertEqual(
            q.value,
            "one",
            "Model value is not 'Test row', got %s" % q.value
        )

    def testQueryLoader(self):
        q = QueryLoaderModel(1)
        self.assertEqual(
            q.id,
            1,
            "Model id is set to 1."
        )
        self.assertEqual(
            q.value,
            "one",
            "Model value is not 'Test row', got %s" % q.value
        )

    def testQueryTyped(self):
        handle = config.handle_factory()
        QueryTypedModel.register_composite('public.test_table', handle)
        model = QueryTypedModel(id=1, handle=handle)
        self.assertEqual(model.value, 'one', 'QueryTypedModel loads correctly')
        cur = model.get(id=2, options={'handle': handle})
        o = cur.fetchone()
        self.assertEqual(o.value, 'two', 'QueryTyped returns correctly')

    def testLazyLoad(self):
        model = SimpleLazyLoaderModel(id=1)
        self.assertEqual(
            model.value,
            "one",
            "Model value is not 'Test row', got %s" % model.value
        )

    def testTypeRegistrationLazy(self):
        handle = config.handle_factory()
        SimpleLazyLoaderModel.register_composite('public.test_table', handle)
        f = FunctionTypedSingle('test_get',['id'], handle=handle)
        model = f(id=1)
        self.assertTrue(isinstance(model, SimpleLazyLoaderModel), 'result row is registered class instance')
        self.assertEqual(
            model.value,
            "one",
            "Model value is not 'Test row', got %s" % model.value
        )

    def testInstanceFunctions(self):
        model = SimpleInstanceModel(id=1)
        cur = model.get()
        row = cur.fetchone()
        self.assertEqual(row[0], 1, 'missing Function argument supplied implicitly')
        model = SimpleUpdateModel(id=1)
        cur = model.update(new_value="Instance function test")
        # Now, the rs should have returned a single row
        self.assertEquals(cur.rowcount,1,"Update returned a single row.")
        row = cur.fetchone()
        self.assertEqual(row[0], True, "Did not successfully update, got %s" % row[0])
        model = SimpleUpdateModel(id=1)
        self.assertEqual(
            model.value,
            "Instance function test",
            "Key was not successfully set, got '%s', expected '%s'" % (
                model.value,
                "Instance function test" )
            )

    def testTypedSet(self):
        handle = config.handle_factory()
        SimpleReturn.register_composite('public.test_table', handle)
        f = FunctionTyped('test',[])
        cur = f()
        for item in cur.fetchall():
            self.assertTrue(isinstance(item, SimpleReturn), 'Each "row" in FunctionTyped cursor result is a single model object')

    def testFunctionCallback(self):
        callback_value = 'value one'
        def test_callback(row):
            print('executing callback on row: {0}'.format(repr(row)))
            model = row[0]
            model.callback_attrib = callback_value
            return row

        handle = config.handle_factory()
        SimpleReturn.register_composite('public.test_table', handle)
        f = FunctionTyped('test',[], callback=test_callback)
        cur = f()
        for model in cur.fetchall():
            self.assertTrue(model.callback_attrib == callback_value, 'Function-initiallized callback was executed on row with id={0}'.format(model.id))
        callback_value = 'value two'
        cur = f(options={'callback': test_callback})
        for model in cur:
            self.assertTrue(model.callback_attrib == callback_value, 'Function call callback was executed on row with id={0}'.format(model.id))


    def testNestedModel(self):
        handle = config.handle_factory()
        SimpleLazyLoaderModel.register_composite('public.test_table', handle)
        NestedModel.register_composite('public.nested', handle)
        model = NestedModel(id=1)
        self.assertTrue(isinstance(model, NestedModel), 'result row is registered class instance')
        self.assertEqual(
            model.value,
            "one",
            "Model value is not 'Test row', got %s" % model.value
        )
        self.assertTrue(isinstance(model.others, list), 'model.others is a list')

    def testProperty(self):
        model = SimpleInstanceModel(id=1)
        self.assertEqual(model.prop, 1, 'Property returns single object')

    def testDynamicModel(self):
        handle = config.handle_factory()
        DynamicModel.register_composite('public.test_table', handle)
        model = DynamicModel()
        self.assertEqual(model.table, SimpleReturn.table, 'table is determined automatically')


class FunctionTest(dbTest):

    def testCreateFunction(self):

        f = Function("test")
        self.failUnless(
            isinstance(f, Function),
            "Isn't a Simpycity function object."
        )


    def testExecuteFunction(self):

        f = Function("test")
        cur = f()
        self.assertEqual(cur.rowcount,3,'Error in execute of function test, expected 3 rows, got %s' % cur.rowcount)


    def testPartialReturnSet(self):
        f = Function("test")
        cur = f(options=dict(columns=['id']))
        self.assertEqual(cur.rowcount,3,"Partial Result Set does not have 3 entries.")

        for row in cur:
            try:
                a = row['value']
                self.fail("Expected no value column, found value column of %s" % row['value'])
            except KeyError,e:
                pass
            except Exception, e:
                self.fail("Failed with exception: %s" %e)


    def testPartialWithArguments(self):
        f = Function("test",['id'])
        rs = f(1,options=dict(columns=['id']))
        self.assertEqual(rs.rowcount,1,"Partial with Arguments returns 1 row.")

        for row in rs:
            try:
                a = row['value']
                self.fail("Expected no value column, found value column of %s" % row['value'])
            except KeyError,e:
                pass
            except Exception, e:
                self.fail("Failed with exception: %s" %e)

class QueryTest(dbTest):

    def testBareQuery(self):

        q = Query("test_table")
        self.failUnless(
            isinstance(q, Query),
            "Return from query creation is of type sql_function."
        )
        cur = q()
        self.assertEqual(cur.rowcount,3,"Bare query result set has %s, expected 3." % cur.rowcount)

    def testWhereQuery(self):

        q = Query("test_table",['id'])
        try:
            cur = q(1)
        except Exception, e:
            self.fail("Failed with exception %s" % e)

        self.assertEqual(cur.rowcount,1,"ResultSet has a single entry")

        row = cur.fetchone()
        self.assertEqual(row['id'],1,'Return row not 1, got %s' % row['id'])
        self.assertEqual(row['value'],'one', 'Return row not "one", got %s' % row['value'])

    def testPartialReturnSet(self):
        q = Query("test_table")
        try:
            cur = q(options=(dict(columns=['id'])))
        except Exception, e:
            self.fail("Failed with exception %s" % e)

        self.assertEqual(cur.rowcount,3,"Partial Result Set has 3 entries, as expected.")

        for row in cur:
            try:
                a = row['value']
                self.fail("Expected no value column, found value column of %s" % row['value'])
            except KeyError,e:
                pass

    def testPartialWithArguments(self):
        f = Function("test",['id'])
        cur = f(1,options=dict(columns=['id']))
        self.assertEqual(cur.rowcount,1,"Partial with Arguments returns 1 row.")

        for row in cur:
            try:
                a = row['value']
                self.fail("Expected no value column, found value column of %s" % row['value'])
            except KeyError,e:
                pass
            except Exception, e:
                self.fail("Failed with exception: %s" %e)

class RawTest(dbTest):

    def testRunQuery(self):
        r = Raw("select * from test_table where id = %s",['id'])
        try:
            cur = r(1)

        except Exception, e:
            self.fail("Failed with exception %s" % e)

        self.assertEqual(cur.rowcount, 1, "ResultSet has single entry.")

        row = cur.fetchone()
        self.assertEqual(row['id'],1,'Return row not 1, got %s' % row['id'])
        self.assertEqual(row['value'],'one', 'Return row not "one", got %s' % row['value'])



class SimpleReturn(SimpleModel):

    table = ['id','value']

class SimpleInstanceModel(SimpleModel):
    table = ['id','value']

    get = Function("test_get",['id'])
    prop = Property('test_constant')

class SimpleLoaderModel(SimpleInstanceModel):
    __load__ = FunctionSingle("test_get",['id'])

class QueryLoaderModel(SimpleInstanceModel):
    __load__ = QuerySingle("test_table",['id'])

class QueryTypedModel(SimpleInstanceModel):
    __load__ = QueryTypedSingle("test_table",['id'])
    loaded_indicator = 'value'
    get = QueryTyped("test_table",['id'])

class SimpleLazyLoaderModel(SimpleReturn):
    __lazyload__ = FunctionSingle("test_get",['id'])
    loaded_indicator = 'value'

class SimpleUpdateModel(SimpleLoaderModel):

    update = Function("update_row",['id','new_value'])

class NestedModel(SimpleLazyLoaderModel):
    table = SimpleLazyLoaderModel.table + ['others']
    __lazyload__ = FunctionSingle("test_nested",['id'])

class DynamicModel(SimpleModel):
    table = []
    pg_type = ('public','test_table')

if __name__ == '__main__':
    setUpModule()
    unittest.main()



from simpycity import NotFoundError
from simpycity.core import FunctionError, meta_query
from simpycity import config as g_config
import psycopg2

def d_out(text):

    if g_config.debug:
        print text

class Construct(object):
    config = None

    def __init__(self, config=None, handle=None, *args,**kwargs):
        """
        A Construct is a basic datatype for Simpycity - basically providing
        a framework that allows for all queries in the Construct to operate
        under a single logical transaction.
        """

        d_out("Construct.__init__: config=%s, handle=%s" % (config, handle))

        if not self.config:
            self.config = config or g_config

        self.init_handle = handle


    @property
    def handle(self):
        if not self.init_handle:
            self.init_handle = self.config.handle_factory(config=self.config)

        return self.init_handle

    def commit(self):
        if self.handle is not None:
            self.handle.commit()
        else:
            raise AttributeError("Cannot call commit without localized handle.")

    def close(self):
        if self.init_handle:
            self.init_handle.close()

    def rollback(self):
        if self.handle is not None:
            self.handle.rollback()
        else:
            raise AttributeError("Cannot call rollback without localized handle.")


class SimpleModel(Construct):

    """
    The basic simple model class.
    Implements the barest minimum of Model functionality to operate.

    The SimpleModel expects a table=[] to declare its columns. However if you leave table an empty list and
    define pg_type as a tuple of two members (schema, type) then
    register_composite() will calculate table with a query to Postgres at runtime.
    SimpleModel expects one of:
    * __load__ to be a FunctionSingle or QuerySingle that loads an instance, based on primary key, from the database.
    * __lazyload__ to be a FunctionSingle that loads an instance from the database, depending on the existance of a value for
      loaded_indicator, which should be a member of table that will only be populated after the instance is fully loaded.

    """

    pg_type = None
    table = []

    def __init__(self, *args, **kwargs):
        """
        Sets up the objects' internal column.
        Tests for the presence of a primary key, and attempts to load a
        description using it.
        """

        if 'config' in kwargs:
            config = kwargs['config']
            del(kwargs['config'])
        else:
            config = None

        if 'handle' in kwargs:
            handle = kwargs['handle']
            del(kwargs['handle'])
        else:
            handle = None

        if args or kwargs:
            d_out("SimpleModel.__init__: Got args of %s" % repr(args))
            d_out("SimpleModel.__init__: Got kwargs of %s" % repr(kwargs))

        #
        # If the type defines a base type, the base instance will be
        # passed in the 'base_' kwarg: merge the base attrs into
        # kwargs and hand over to superclass.
        #
        SimpleModel.merge_base_attrs(kwargs)

        #
        # Initialize table attrs by copying from keyword arguments.
        #
        # This makes sure the objects not loaded from DB has at least
        # the None values in attrs and access to them doesn't raise.
        #
        if hasattr(self, 'table'):
            for name in self.table:
                self.__dict__[name] = kwargs.get(name, None)

        # should automatically pick up config= and handle=
        super(SimpleModel, self).__init__(config, handle, *args, **kwargs)

        if hasattr(self, 'loaded_indicator') and self.__dict__.get(object.__getattribute__(self, 'loaded_indicator')) is not None:
            self._loaded = True
        else:
            self._loaded = False

        # config and handle have been dealt with, now.
        if not self._loaded and hasattr(self, '__load__'):
            if args or kwargs:
                self.__load_by_key__(*args, **kwargs)


    def __load_by_key__(self, *args, **kwargs):
        """
        Private method.
        Using the primary key from __init__, load the database row by primary
        key.
        """
        # check for an __load__ function

        if 'options' in kwargs:
            opts = kwargs['options']
            del(kwargs['options'])
        else:
            opts = {}

        opts['handle'] = self.handle

        kwargs['options'] = opts

        row = None
        try:
            row = self.__load__(*args, **kwargs)
        except psycopg2.InternalError, e:
            d_out("pgerror=%s pgcode=%s diag=%s" % (e.pgerror, e.pgcode, e.diag))
            if not (e.pgcode == 'P0002'): # no_data_found
                raise # as InternalError

        if row is None:
            raise NotFoundError()

        d_out("SimpleModel.__load_by_key__: rs: %s" % row)
        if isinstance(row, psycopg2.extras.DictRow):
            loaded_attrs = dict(row)
        elif isinstance(row, SimpleModel):
            #python 2.7+:
            #loaded_attrs = {_ for _ in rs.__dict__.iteritems() if _[0] in self.table}
            loaded_attrs = {}
            for col in self.table:
                loaded_attrs[col] = row.__dict__[col]
        else:
            raise Exception("row is type {0}".format(type(row)))
        SimpleModel.merge_base_attrs(loaded_attrs)
        for item in self.table:
            d_out("SimpleModel.__load_by_key__: %s during load is %s" % (item, repr(loaded_attrs[item])))
            self.__dict__[item] = loaded_attrs[item]
        self._loaded = True
        d_out("SimpleModel.__load_by_key__: self.__dict__ is %s" % self.__dict__)

    def __getattribute__(self,name):

        """
        Private method.

        This function tests all attributes of the SimpleModel if they are
        instances of meta_query.
        If they are a meta_query, columns from the Model are mapped
        to the InstanceMethods' arguments, as appropriate.
        """

        attr = object.__getattribute__(self,name)
        if name == '__load__':
            d_out("skipping: %s" % name)
            return attr

        if attr is None and name in object.__getattribute__(self, 'table') and not object.__getattribute__(self,'_loaded'):
            should_lazyload = hasattr(self,'__lazyload__')
        else:
            should_lazyload = False

        if should_lazyload:
            d_out("lazyloading {0} on {1}".format(self.__class__, name))
            attrs = object.__getattribute__(self, '__dict__')
            attrs['_loaded'] = True
            rs = self.__lazyload__(options={'handle':self.handle})
            if not rs:
                raise NotFoundError("__lazyload__ returned: {0}".format(rs))
            if isinstance(rs, psycopg2.extras.DictRow):
                loaded_attrs = dict(rs)
            elif isinstance(rs, SimpleModel):
                #python 2.7+:
                #loaded_attrs = {_ for _ in rs.__dict__.iteritems() if _[0] in self.table}
                loaded_attrs = {}
                for col in self.table:
                    loaded_attrs[col] = rs.__dict__[col]
            else:
                raise Exception("__lazyload__ returned unexpected type: {0}".format(type(rs)))
            SimpleModel.merge_base_attrs(loaded_attrs)
            attrs.update(loaded_attrs)
            return attrs[name]

        if isinstance(attr, meta_query):

            d_out("SimpleModel.__getattribute__: Found meta_query %s" % name)
            def instance(*args,**kwargs):

                if args:
                    raise FunctionError("This function can only take keyword arguments.")
                my_args = kwargs.copy()
                d_out("SimpleModel.__getattribute__ InstanceMethod: kwargs: %s" % repr(kwargs))
                d_out("SimpleModel.__getattribute__ InstanceMethod: self.__dict__: %s" % self.__dict__)
                for arg in attr.args:
                    d_out("SimpleModel.__getattribute__ InstanceMethod: checking arg %s" % arg)
                    if arg not in kwargs:
                        d_out("not in my_args")
                        if hasattr(self, arg):
                            d_out("SimpleModel.__getattribute__ InstanceMethod: found %s in col.." % arg)
                            my_args[arg] = getattr(self, arg)
                        else:
                            my_args[arg] = None

                if 'options' not in kwargs:
                    d_out("SimpleModel.__getattribute__ InstanceMethod: Didn't find options.")
                    my_args['options'] = {}

                # pass self to the query object
                my_args['options']['model'] = self

                d_out("SimpleModel.__getattribute__: InstanceMethod: Setting handle.")
                my_args['options']['handle'] = self.handle
                rs = attr(*args, **my_args)
                d_out("SimpleModel.__getattribute__: InstanceMethod: model :{model} attrib name: {name} (attrib value: {value}) constructor returned {rs}".format(model=self, name=name, value=attr, rs=rs))
                return rs

            if attr.is_property:
                return instance()
            else:
                return instance

        else:
            return attr

    def set(self,col,val):
        self.__dict__[col] = val


    def save(self):

        """Performs the __save__ method, if it has been declared.
        If not, this function raises a CannotSave exception.
        .save() does *not* implicitly commit the model.
        To commit, it must be done manually."""

        if hasattr(self, "__save__"):

            args = self.__save__.args()
            my_args = {}
            for arg in args:
                my_args[arg] = self.__dict__.get(arg, None)
            rs = self.__save__(**my_args).fetchone()

            for arg in self.table:
                if arg in rs:
                    self.__dict__[arg] = rs[arg]
        else:
#            from simpycity import CannotSave
            raise NotImplementedError("Cannot save without __save__ declaration.")

    @classmethod
    def register_composite(cls, name, handle=None, factory=None):
        """
        Maps a Postgresql type to this class.  If the class's table attribute
        is empty, and the class has an attribute pg_type of tuple (schema, type),
        it is calculated and set by querying Postgres. Register inherited/inheriting
        classes in heirarchical order.
        Every time a SQL function returns a registered type (including array
        elements and individual columns, recursively), this class
        will be instantiated automatically.

        The object attributes will be passed to the provided callable in
        a form of keyword arguments.

        :param name: the name of a PostgreSQL composite type, e.g. created using
            the |CREATE TYPE|_ command
        :param handle: Simpycity handle
        :param factory: if specified it should be a `psycopg2.extras.CompositeCaster` subclass: use
            it to :ref:`customize how to cast composite types <custom-composite>`
        :return: the registered `CompositeCaster` or *factory* instance
            responsible for the conversion
        """
        class CustomCompositeCaster(psycopg2.extras.CompositeCaster):

            def make(self, values):
                d_out("CustomCompositeCaster.make: cls={0} values={1}".format(repr(cls), repr(values)))
                return cls(**dict(zip(self.attnames, values)))

        PG_TYPE_SQL = """SELECT array_agg(attname)
FROM
    (
        SELECT attname
        FROM
            pg_type t
            JOIN pg_namespace ns ON typnamespace = ns.oid
            JOIN pg_attribute a ON attrelid = typrelid
        WHERE nspname = %s AND typname = %s
            AND attnum > 0 AND NOT attisdropped
        ORDER BY attnum
    ) sub;"""
        if handle is None:
            handle = g_config.handle_factory()
        d_out("SimpleModel.register_composite: before: table for {0} is {1}".format(repr(cls.pg_type), cls.table))
        if cls.pg_type is not None:
            super_table = cls.__mro__[1].table if hasattr(cls.__mro__[1], 'table') else []
            if cls.table == [] or cls.table is super_table:
                cursor = handle.cursor()
                cursor.execute(PG_TYPE_SQL, cls.pg_type)
                row = cursor.fetchone()
                d_out("SimpleModel.register_composite: row={0}".format(row))
                row[0] = [_ for _ in row[0] if _ != 'base_']
                cls.table = cls.table + row[0]
                d_out("SimpleModel.register_composite: after: table for {0} is {1}".format(repr(cls.pg_type), cls.table))
        if factory is None:
            factory = CustomCompositeCaster
        return psycopg2.extras.register_composite(
            name,
            handle.conn,
            globally=True,  # in case of reconnects
            factory=factory
        )

    @staticmethod
    def merge_base_attrs(attrs):
        """
        If one of the attrs is named "base_", assume that attribute is an instance of SimpleModel mapped on a Postgresql
        composite type, and that the base_ instance is of a superclass of this class. Expand the attributes of the
        base_ type and assign to class attributes.

        psycopg2's type casting uses namedtuple() and that forbids a
        name to start with underscore, so we end it with _ instead
        """
        base = attrs.pop('base_', None)
        if base:
            d_out("SimpleModel.merge_base_attrs: base.table={0}".format(base.table))
            for name in base.table:
                attrs[name] = base.__dict__[name]



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='Simpycity',
    version='2.0.0',
    description='A database-respecting object-relational mapper for PostgreSQL.',
    long_description=long_description,
    author='Command Prompt, Inc.',
    author_email='support@commandprompt.com',
    url='https://github.com/commandprompt/Simpycity',
    download_url='https://github.com/commandprompt/Simpycity/releases/tag/2.0.0',
    install_requires=[
        "psycopg2>=2.5"
    ],
    packages=find_packages(),
    test_suite='nose.collector',
    license='LGPL',
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Database",
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Natural Language :: English'
    ],
    keywords='orm postgresql',
)




"""
    COPYRIGHT 2009-2016 Command Prompt, Inc.
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import psycopg2
from simpycity import config, ProgrammingError
import simpycity.handle

def d_out(text):

    if config.debug:
        print text


class meta_query(object):

    """
    Base object for sql query-like objects For internal use only.
    """

    def __init__(self, name, args=[], handle=None, callback=None):

        """

         :name    Sets the base name of the query. How this is used will be
                    declared in the implementing subclass. For instance, in
                    the Function subclass, name is the name of the stored
                    procedure itself. In the case of Raw, it is the entire
                    query.
         :args    A base empty list, declaring the arguments, if any, that
                    this query requires.

         :handle   Instead of creating a new handle to run this query,
                    utilize the provided handle. This is handled implicitly
                    by the SimpleModel.

         :callback  Each row returned by the cursor will be passed to this function, which must return the row.
                      The function is applied to the psycopg row as returned by psycopg, before any other Simpycity handling.
                      Can be overriden on call-to-call basis via options= of the  __call__ methond.
        """

        self.query_base = name
        self.args = args
        self.__attr__ = {}

        self.__attr__['handle'] = handle
        self.__attr__['callback'] = callback
        self.cursor_factory = simpycity.handle.Cursor

    def __call__(self, *in_args, **in_kwargs):

        """
        Call function for a meta_query/sql_function-style object.

        Calling the function in this method causes the core SQL to be run and
        a ResultSet returned.

        :param options: A dict, which responds to the following
        keys:
        * columns: Alters what columns are selected by the query.
        * handle: Overrides the instance handle with a customized version.
        * callback: Override the instance callback with a customized version.
        :return psycopg2 cursor
        """

        d_out("meta_query.__call__: query is %s" % self.query_base)

        d_out("meta_query.__call__: Got args %s" % in_kwargs)

        opts = in_kwargs.pop('options', None)
        if opts:
            # we are going to pop out the options we handle ourselves,
            # then pass any extra options down: make a copy to avoid
            # modifying the passed options dict
            opts = opts.copy()
        else:
            opts = {}

        columns = opts.pop('columns', [])
        handle = opts.pop('handle', self.__attr__['handle'])
        callback = opts.pop('callback', self.__attr__['callback'])

        if len(columns) >= 1:
            # we are limiting the return type.
            # Eventually, we'll check it against the return object.
            # Until then, we just assume the user knows what they're
            # doing.
            cols = ",".join([x for x in columns])
            d_out("meta_query.__call__: Called with column limiters: %s" %cols)


        else:
            cols = "*"

        d_out("meta_query.__call__: Requires args: %s" % len(self.args))
        d_out("meta_query.__call__: Got args: %i" % (len(in_kwargs) + len(in_args)))

        d_out("in_args: %s" % str(in_args))
        d_out("in_kwargs: %s" % str(in_kwargs))

        # If we were called with arguments
        if len(in_args) >= 1:

            # Tests if the number of positional arguments + the number of
            # keyword arguments is less than the number of arguments this
            # instance was declared to require.

            if len(in_args) < len(self.args) \
                and len(in_kwargs) < len(self.args) \
                and len(in_kwargs) + len(in_args) < len(self.args):
                raise Exception("Insufficient arguments: Expected %s, got %s"
                                % (len(self.args), len(in_args)+len(in_kwargs)) )

            # Tests if the number of positional arguments + the number of
            # keyword arguments is GREATER than the number of arguments this
            # instance was declared to require.

            if len(in_args) > len(self.args) \
                or len(in_kwargs) > len(self.args) \
                or len(in_kwargs) + len(in_args) > len(self.args):
                raise Exception("Too many arguments: Expected %s, got %s" %
                                (len(self.args), len(in_args)+len(in_kwargs)))

        # Create a fixed-length array equal to the number of arguments
        # this instance requires.

        call_list = ['' for x in xrange(len(self.args))]

        if in_kwargs:
            # Map the incoming keyword args positionally, based on the
            # position of argument names in the core argument list.

            for arg in in_kwargs.iterkeys():
                try:
                    call_list[ self.args.index(arg) ] = in_kwargs[arg]
                except ValueError:
                    raise Exception("Unknown keyword argument passed: %s" % arg)

        for index,arg in enumerate(in_args):
            call_list[index] = arg
        d_out("meta_query.__call__: Handle is %s" % handle)
        d_out("meta_query.__call__: callback is %s" % callback)
        cur = self.__execute__(cols, call_list, handle, callback, extra_opt=opts)
        d_out("meta_query.__call__: returning cur of %s" % cur)
        return cur


    @property
    def is_property(self):
        """
        Override this for special handling in SimpleModel.__getattribute__
        If True, __call__ will be executed if a reference is made to a SimpleModel attribute
        that is this instance.
        """
        return False

    def form_query(self, columns, options={}):
        """Subclass function to create the query based on the columns
        provided at instance time.
        :return sql string
        """
        pass

    def handle(self, handle):

        """
        Permanently resets the handle.
        """
        self.__attr__['handle'] = handle

    def __execute__(self, columns, call_list, handle=None, callback=None, extra_opt={}):
        '''
        Runs the stored query in a psycopg2 cursor based on the arguments provided to
        __call__.
        If the instance handle is None and also the handle parameter is none, a handle
        is created from config.handle_factory().
        :param extra_opt: a dict passed to form_query
        :return psycopg2 cursor
        '''

        query = self.form_query(columns, options=extra_opt)

        d_out("meta_query __execute__: Handle is %s" % handle)

        if handle is None:
            if self.__attr__['handle'] is None:
                d_out("meta_query.__execute__: Did not find handle, creating new.. ")
                handle = config.handle_factory()
                self.__attr__['handle'] = handle
                d_out("meta_query.__execute__: Handle is %s" % self.__attr__['handle'])
            else:
                d_out("meta_query.__execute__: Found object handle.. ")
                handle = self.__attr__['handle']

        cursor = handle.cursor(cursor_factory=self.cursor_factory, callback=callback)
        d_out("meta_query.__execute__: Cursor is %s" % cursor)
        d_out("meta_query.__execute__: Query: %s" % ( query ) )
        d_out("meta_query.__execute__: Call List: %s" % ( call_list ) )

        try:
            cursor.execute(query, call_list)

        except psycopg2.OperationalError as e:
            # retry query on stale connection error
            d_out("OperationalError: %s" % e)

            cursor = handle.cursor()
            cursor.execute(query, call_list)

        return cursor

    def commit(self):

        """Commits the query, using the internal self.handle."""

        if self.__attr__['handle'] is not None:
            self.__attr__['handle'].commit()
        else:
            raise Exception("Could not commit, no handle found.")

    def rollback(self):

        """Rolls back the query, using the internal self.handle."""

        if self.__attr__['handle'] is not None:
            self.__attr__['handle'].rollback()
        else:
            raise Exception("Could not rollback, no handle found.")


class Function(meta_query):
    """
    Abstract a Postgresql function.
    """
    def __init__(self, *args, **kwargs):
        """
        :param name: sql function name
        :param args: list of sql argument names
        :param direct: perform direct query "SELECT func(args...)" when True,
                 vs. "SELECT * FROM func(args...)" when False (default.)
        """
        self.direct = kwargs.pop('direct', False)
        super(Function, self).__init__(*args, **kwargs)

    def form_query(self, columns, options={}):
        """
        :param columns: literal sql string for list of columns
        :param options: dict supporting a single key "direct" as in the constructor
        :return sql string
        """
        from_cl = 'FROM'

        direct = options.get('direct', self.direct)
        if direct:
            if columns != '*':
                raise ProgrammingError("Column lists cannot be specified for a direct function call.")
            columns = ''
            from_cl = ''

        if len(self.args) >= 1:
            replace = ['%s' for x in xrange(len(self.args))]
            func = "%s(" % self.query_base + ",".join(replace) + ")"
        else:
            func = "%s()" % self.query_base

        return "SELECT %s %s %s" % (columns, from_cl, func)


class FunctionSingle(Function):
    """
    A Postgresql function that returns a single value.
    """
    def __call__(self, *in_args, **in_kwargs):
        cursor = super(FunctionSingle, self).__call__(*in_args, **in_kwargs)
        if cursor.rowcount <> 1:
            raise Exception("Expect only a single row")
        row = cursor.fetchone()
        return row


class FunctionTyped(Function):
    """A Postgresql function that returns row(s) having only a single (typically composite) column"""

    def __init__(self, *args, **kwargs):
        kwargs.pop('direct',None)
        super(FunctionTyped, self).__init__(*args, **kwargs)
        self.direct = True
        self.cursor_factory = simpycity.handle.TypedCursor


class FunctionTypedSingle(FunctionSingle, FunctionTyped):
    """A Postgresql function that returns a single row having a single (typically composite) column"""
    pass

class Property(FunctionTypedSingle):
    """
    Enjoys special handling in SimpleModel.__getattribute__
    When a SimpleModel attribute references an instance of this class, its __call__
    attribute is executed immediately and the result returned. That means in
    __init__, args must be empty, or the args must be mappable by name to the model's table attribute.
    """

    @property
    def is_property(self):
        return True


class Raw(meta_query):
    """
    Execute arbitrary sql.
    """
    def __init__(self, name, args=[], handle=None, callback=None):
        """
        :param name: The raw sql
        :param args: noop
        :param handle: see superclass
        :param callback: see superclass
        """
        super(Raw, self).__init__(name, args, handle, callback)

    def form_query(self, columns, options={}):
        #TODO: use args for parameterized query
        return self.query_base

class Query(meta_query):
    """
    select query access to a Postgresql table or view.
    """
    def __init__(self, name, args=[], handle=None, callback=None):
        """
        :param name: table or view name
        :param args: list of column names used in sql WHERE clause
        :param handle: see superclass
        :param callback: see superclass
        """
        super(Query, self).__init__(name, args, handle, callback)
        self.direct = False

    def form_query(self, columns, options={}):
        where_list = None
        if len(self.args) >= 1:
            where_list = [x+"=%s" for x in self.args]

        if self.direct:
            if columns != '*':
                raise ProgrammingError("Column lists cannot be specified for a typed query call.")
            columns = 'row(t.*)::{base}'.format(base=self.query_base)

        query = "SELECT {columns} FROM {base} t".format(columns=columns, base=self.query_base)
        if where_list:
            query += " WHERE " + " AND ".join(where_list)

        return query

class QuerySingle(Query):
    """
    A select query on a table or view that returns a single row
    """
    def __call__(self, *in_args, **in_kwargs):
        cursor = super(QuerySingle, self).__call__(*in_args, **in_kwargs)
        if cursor.rowcount <> 1:
            raise Exception("Expect only a single row")
        row = cursor.fetchone()
        return row


class QueryTyped(Query):
    """A select query on a table or view that returns composite row values"""

    def __init__(self, *args, **kwargs):
        kwargs.pop('direct',None)
        super(QueryTyped, self).__init__(*args, **kwargs)
        self.direct = True
        self.cursor_factory = simpycity.handle.TypedCursor


class QueryTypedSingle(QuerySingle, QueryTyped):
    """A select query on a table or view that returns a single composite row value"""
    pass

class FunctionError(BaseException):
    """
    Bare exception, used for naming purposes only.
    """
    pass

class ProceduralException(BaseException):

    pass




port=None
host=None
database=None
user=None
password=None
debug=False


def dsn():
    """
    Return a libpq connection string using the variables defined in this file.
    """
    configs = {'host': host,
               'port': port,
               'dbname': database,
               'user': user,
               'password': password}
    return ' '.join(['{0}={1}'.format(_[0], _[1]) for _ in configs.iteritems() if _[1] is not None and _[1] != ''])

def handle_factory(*args, **kwargs):
    """
    Simpycity calls this to get a handle when it needs one.
    It may be useful to redefine this function with one that returns an existing instance of Handle,
    so that your application only opens one -- or a controlled number of -- database connections.
    """
    from handle import Handle
    return Handle(*args, **kwargs)



"""
    COPYRIGHT 2008-2016 Command Prompt, Inc.
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lessor General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import psycopg2

DataError = psycopg2.DataError
DatabaseError = psycopg2.DatabaseError
IntegrityError = psycopg2.IntegrityError
InternalError = psycopg2.InternalError
InterfaceError = psycopg2.InterfaceError
OperationalError = psycopg2.OperationalError
ProgrammingError = psycopg2.ProgrammingError
NotSupportedError = psycopg2.NotSupportedError


class PermissionError(InternalError):
    pass

class UnknownUserError(PermissionError):
    pass

class NotFoundError(InternalError):
    pass

class CannotSave(InternalError):
    pass

